#ifndef PARTSUM_H
#define PARTSUM_H
#include "globdef.h"

using namespace std;

// =====================================================
//   TRANSLATE FUNCTION

/*	The translate(DNAstrand) function will translate
	the chars in the strand to the values 0,1,2,3
	meaning A,C,G,T for DNA, and 4,5,6,7 meaning
	A,C,G,U for RNA.
	This will allow for a quicker calculation of the
	partition sum. These are dynamically allocated,
	to allow any strand length.
*/
// =====================================================

void translate(string strand, short int * numstr)
{
    int i; // counter
    int strlength = strand.length();

    for(i=0;i<strlength;i++)
    {
	numstr[i] = 8;

	if(strand.at(i) == 'A') numstr[i] = 0;
	else if(strand.at(i) == 'C') numstr[i] = 1;
	else if(strand.at(i) == 'G') numstr[i] = 2;
	else if(strand.at(i) == 'T') numstr[i] = 3;
	else if(strand.at(i) == 'B') numstr[i] = 4;
	else if(strand.at(i) == 'D') numstr[i] = 5;
	else if(strand.at(i) == 'H') numstr[i] = 6;
	else if(strand.at(i) == 'U') numstr[i] = 7;
	else numstr[i] = 15;

	if(numstr[i] == 8)
	{
	    cout << "FATAL ERROR!\n Invalid character '"<< strand.at(i) << "' encountered!\n\n";
	    exit(1);
	}
    }
}
// =====================================================
//   end TRANSLATE FUNCTION
// =====================================================

// =====================================================
//   PARTITION FUNCTION

/*  The Partition Function will use the known values
    for the pair binding energies to determine the
    total partition sum for a given configuration.
    To determine the Partition Sum, a double sum has to
    be evaluated.
    Function variables:
    * strand: A strand of numbers 0-7 to represent RNA
    * strlength: The length of the Dyn. All. strand
    * expbdg: binding energy values
    * Partial_PS: array for storage of results
    * Partial_PSbow: array for storage of results with start-end arc

    Characteristics:
    This is the most intuitive approach. All partition
    sums are calculated, using old results.

    Method summary:
    This function consists of a triple loop. The
    outermost loop runs over the length j of the partial
    strand. The second runs over the position i. Then,
    for each partial partition sum p(i,j),
    initialization is done: p(i,j) = p(i,j-1) +
    pbow(i,j-1). All possible arcs connecting to the
    j-th nucleotide are then found and added to the
    total. In this manner, all partial partition sums
    are computed in an order-N^3 algorithm.

    Returned value:
    The returned value is the sum of all possible
    configurations, hence it consists of Partial_PS, all
    configurations without the start-end bow, and
    Partial_PSbow, which (if nonzero) contains the
    configurations which had the first and last
    nucleotide connected to each other.
*/
// =====================================================

double partsum(short int * strand, int strlength, double ** expbdg, double ** Partial_PS, double ** Partial_PSbow, double * extrapar)
{
    if(MINBOW >= strlength-2) return 1; // There can be no possible connection. But
    		// ... if we continue, terrible things happen, so return 1 right now.
    int i, j, k;  // defining the necessary variables for the triple loop
    double betadeltag; // To keep notation short in calculations
    double prebow = 0; // The partition sum ahead of the bow, which may have a bow i->k-1
    double innerbow; // The partition sum inside AND including the bow i+k->i+j

    double innerfactor, outerfactor; // terminal AU pair effect

    int type;

    bool combined = 0; // Combined?
    if(strand[1+strlength/2] == 15) combined = 1; // Only if there's an X in the middle of the strand, which represents two (equal length-)strands glued together.
  
    // Initialize the result arrays.
    for(i = 0; i<strlength; i++)
	for(j = 0; j<strlength+1; j++)
	{
	    Partial_PSbow[i][j] = 0;
	}

    for(i = 0; i<strlength; i++)
	for(j = 0; j<MINBOW+3; j++)
	    Partial_PS[i][j] = 1; // Fill the first few lines of the array

    for (j=MINBOW+3; j<strlength; j++) // j is the bowlength
    {
	for (i=0; i < (strlength - j); i++) // i is the starting position of the i-j partition sum under consideration
	{
	    Partial_PS[i][j] = Partial_PS[i][j-1];	// Start with the No-Bow case, there is NO i->j bow.
	    type = strand[i]/4 + strand[i+j-1]/4;
	    if(strand[i] == 0 || strand[i] == 3 || strand[i] == 4 || strand[i] == 7)
	    {
		    outerfactor = extrapar[14+type];
	    }
	    else
	    {
		outerfactor = 1;
	    }
	    Partial_PS[i][j] += Partial_PSbow[i][j-1]*outerfactor; // Now, we DO want to allow this bow in the partition sum
	    for (k=0; k<j-2-MINBOW; k++)		// Consider a bow i+k->i+j and i+k+1->i+j-1
	    {
// To fit this on a page, the indent was shifted to the left!

// First, check if the bow is favorable, if nucleotides are not complementary, quit current loop.
// To do this, consider only those combinations whose indices sum up to 3 or 7 or 11.
// Since dAdT = 0+3, dCdG = 1+2, dArU = 0+7, dCrG = 1+6, rArU = 4+7, et cetera,
// and no illegal combination of two elements of the
// set {0,1,2,3,4,5,6,7} gives 3 modulo 4. (Solutions are 0+3,1+2,0+7,1+6,2+5,3+4,4+7,5+6
// which correspond to dAdT,dCdG,dArU,dCrG,dGrC,dTrA,rArU,rCrG. Thus, we have a one-to-one
// mapping between pairs of numbers <8 and NN pairs, designed such that every complementary
// pair sums up to 3 mod 4, while every noncomplementary pair sums up to something else.
		
// Don't do anything unless there is a bow
if( ((strand[i+k] + strand[i+j]) % 4) != 3 || ((strand[i+k+1] + strand[i+j-1]) % 4) != 3)
    continue; // some things don't match
if( (strand[i+k] + strand[i+j]) > 14 || strand[i+k+1] + strand[i+j-1] > 14)
    continue; // we're having X's here, forbidden

// Calculate the value of the exponent for this bow
betadeltag = expbdg[strand[i+k]][strand[i+j-1]];

if(k>0) // Prebow (value of partition sum outside the bow under consideration) is only used for k>0
{
    type = strand[i]/4 + strand[i+k-1]/4;
    if(strand[i+k-1] == 0 || strand[i+k-1] == 3 || strand[i+k-1] == 4 || strand[i+k-1] == 7)
    {
	    outerfactor = extrapar[14+type];
    }
    else outerfactor = 1;
    // Prebow consists of the partition sum (bow or no-bow i->i+k-1) from the start (i) to the bow (i+k-1)
    if(k>3) prebow = Partial_PS[i][k-1] + Partial_PSbow[i][k-1]*outerfactor; // The TOTAL partition sum for i up to i+k-1
    else prebow = 1;
}

type = strand[i+k]/4 + strand[i+j-1]/4;
if(strand[i+k+1] == 0 || strand[i+k+1] == 3 || strand[i+k+1] == 4 || strand[i+k+1] == 7) innerfactor = extrapar[14+type];
else innerfactor = 1;
// Inner bow consists of (Bow energy) x (Partition sum of i+k+2->i+j-2 with NO bow between the ends)
// 		    +	 (Bow energy) x (Partition sum of i+k+1->i+j-1 with FORCED bow between the ends)
if(k<j-3)
{
    if(Partial_PS[i+k+2][j-k-4] == 1 && (combined == 0 || i+j-1 < strlength/2 || i+k > strlength/2)) // HAIRPIN
    {
	innerfactor *= extrapar[4 + 7*(strand[i+k]/4)];
    }
    else if(combined == 0 || i+j-1 < strlength/2 || i+k > strlength/2) // BULGE (disabled)
    {
//	cout << "Detected bulge, since combined = " << combined << "while start = " << i+k << " and end = " << i+j-2 << endl;
	//innerfactor *= extrapar[ 17 + type ];
    }
//    else if(Partial_PS[i+k+2][j-k-4] > 1.0001) innerfactor *= extrapar[ 17 + type ]; // This is a =O= type bulge, but a bulge nonetheless.
    // Else, we have crossed the middle of a glued strand and no bulge or loop contribution appears.
    innerbow = betadeltag*(Partial_PS[i+k+2][j-k-4]*innerfactor+Partial_PSbow[i+k+1][j-k-2]);
}
else {cout << "MINBOW seems to be equal to 0. This is not allowed for the current program. Terminating. " << endl; exit(1);}

if(k==0) Partial_PSbow[i+k][j-k] = innerbow; // if k=0, haven't seen this bow before, so we should remember it for future use
if(k>0)
{
    type = strand[i+k]/4 + strand[i+j-1]/4;
    if(strand[i+k] == 0 || strand[i+k] == 3 || strand[i+k] == 4 || strand[i+k] == 7)
    {
	if(i+k<2 && i+j > strlength-3)
	    outerfactor = extrapar[20+type]; // We are near the ends of the strand, so use the termAU
	else
	    outerfactor = extrapar[14+type]; // We are not near the ends of the strand, so use intAU for internal AU-pair initiation
    }
    else
    {
	    outerfactor = 1;
    }
    Partial_PS[i][j] += innerbow*prebow*outerfactor;
}
	    }
	}
    }
    type = strand[0]/4 + strand[j-1]/4;
    if(strand[0] == 0 || strand[0] == 3 || strand[0] == 4 || strand[0] == 7) outerfactor = extrapar[20+type];
    else outerfactor = 1;
    double partsum = Partial_PS[0][strlength-1] + Partial_PSbow[0][strlength-1]*outerfactor;

  return partsum;
}

#endif
