#ifndef SEQUENCE_H
#define SEQUENCE_H
/* This header file initializes the sequences and their experimental values for deltaG, the binding energy.
   To construct the list of sequences, a class is declared, containing all relevant data for each sequence. */
#include<string>
#include<iostream>
#include<math.h>
#include "globdef.h"

using namespace std;

class Sequence
{
private:
    int seqID;
    string seq; // The given sequence (see Xia et al. for a list of sequences)
    string rcseq; // The matching sequence
    string combseq; // The two sequences combined into one for rainbow diagrams
    short int * numstrand[3];
    double deltaG; // the total binding energy of two strands from the experiment
    double deltaGref; // the total binding energy predicted by Allawi and SantaLucia (1997, Biochemistry 36, 10581)
    double deltaH; // The enthalpy for the strand
    double deltaS; // The entropy
    double melttemp; // The melting temperature at total strand concentration 1x10^-4 M.
    int termAUcount; // the number of terminal AU pairs
    bool symmcheck; // whether the strand is its own complement
public:
    void calcpartsum(double &strand, double &rcstrand, double &combstrand, double ** expbdg, double ** Partial_PS, double ** Partial_PSbow, double * extrapar);

    void remake(void)
    {
	numstrand[0] = new short int[seq.length()];
	numstrand[1] = new short int[rcseq.length()];
	numstrand[2] = new short int[combseq.length()];
	translate(seq,numstrand[0]);
	translate(rcseq,numstrand[1]);
	translate(combseq,numstrand[2]);
    }

    void printstr(void)
    {
	cout << "Contents of combined strand: ";
    for(unsigned int n = 0; n<combseq.length(); n++) cout << numstrand[2][n] << ",";
    cout << endl;
    }
    Sequence() // Constructor with no arguments
    {
	seq = rcseq = combseq = "XXXXXXXXXXXXXXXXXXX";
	deltaG = 0;
	numstrand[0] = new short int[seq.length()];
	numstrand[1] = new short int[rcseq.length()];
	numstrand[2] = new short int[combseq.length()];
	translate(seq,numstrand[0]);
	translate(rcseq,numstrand[1]);
	translate(combseq,numstrand[2]);
    }

    Sequence(const Sequence &sequence)
    {
	cout << "This is the Copy Constructor." << endl;
	seqID = sequence.seqID;
	deltaG = sequence.deltaG;
	deltaGref = sequence.deltaGref;
	deltaH = sequence.deltaH;
	deltaS = sequence.deltaS;
	melttemp = sequence.melttemp;
	termAUcount = sequence.termAUcount;
	symmcheck = sequence.symmcheck;
	seq = sequence.seq;
	rcseq = sequence.rcseq;
	combseq = sequence.combseq;
	numstrand[0] = new short int[seq.length()];
	numstrand[1] = new short int[rcseq.length()];
	numstrand[2] = new short int[combseq.length()];
	translate(seq,numstrand[0]);
	translate(rcseq,numstrand[1]);
	translate(combseq,numstrand[2]);
    }

    Sequence(int id, string str, double energ, double energRef, double enthalpy, double entropy) // DNA/RNA sequence constructor
    {
	seqID = id-1; // The list starts at 1, but C++ arrays start with a zeroth element. Thus, let us prevent confusion.
	deltaG = energ;
	deltaGref = energRef;
	deltaH = enthalpy;
	deltaS = entropy;
	melttemp = deltaH/(deltaS + 1.987*log(0.000025)); // R = 1.987 cal/(K mol), gas constant. Using eq(1) from Sugimoto, 1995.
        seq = str;
	string strand = str;
	string rcstrand = str;
        for(unsigned int i = 0; i<strand.length(); i++)
        {
            if(strand.at(i) == 'B') rcstrand.replace(i,1,"T",1);
            else if(strand.at(i) == 'D') rcstrand.replace(i,1,"G",1);
            else if(strand.at(i) == 'H') rcstrand.replace(i,1,"C",1);
            else if(strand.at(i) == 'U') rcstrand.replace(i,1,"A",1);
        }
	// Note that rcseq now contains the complementary RNA strand! Only, the order is still incorrect (i.e. reversed). Thus, let us reorder:
        string temp;
        int length;
        char tmpchar;
        string charstr;
        length = strand.length();
        temp = rcstrand.at(length-1);
        for(int i = 1; i<length; i++)
        {
            tmpchar = rcstrand.at(length-i-1);
            charstr = tmpchar;
       	    temp.append(charstr);
        }
        rcseq = temp; // Invert direction, now it is in 5'-3' direction
	symmcheck = 1;
	for(int i = 0; i<length; i++) symmcheck *= (seq.at(i) == rcseq.at(i)); // This is the symmetry check
        combseq = seq;
        combseq.append("XXXXX");
        combseq.append(rcseq);
	termAUcount = 0;
	if(seq.at(0)=='A' || seq.at(0)=='B' || seq.at(0)=='T' || seq.at(0)=='U') termAUcount++;
	if(rcseq.at(0)=='A' || rcseq.at(0)=='B' || rcseq.at(0)=='T' || rcseq.at(0)=='U') termAUcount++;
	numstrand[0] = new short int[seq.length()];
	numstrand[1] = new short int[rcseq.length()];
	numstrand[2] = new short int[combseq.length()];
	translate(seq,numstrand[0]);
	translate(rcseq,numstrand[1]);
	translate(combseq,numstrand[2]);
    }

    Sequence(int id, string str, double energ, double energRef, double dh, double ds, bool dummy) // This is the DNA/DNA Sequence constructor
    {
	seqID = id-1; // The list starts at 1, but C++ arrays start with a zeroth element. This seems more intuitive, even if it may cause confusion.
	seqID += int(dummy); // prevent an 'unused' warning, dummy is used only to distinguish RNA/RNA from DNA/DNA constructor and has no purpose other than that.
	deltaG = energ;
	deltaGref = energRef;
	deltaH = dh;
	deltaS = ds;
        seq = str;
	string strand = str;
	string rcstrand = str;
        for(unsigned int i = 0; i<strand.length(); i++)
        {
            if(strand.at(i) == 'A') rcstrand.replace(i,1,"T",1);
            else if(strand.at(i) == 'C') rcstrand.replace(i,1,"G",1);
            else if(strand.at(i) == 'G') rcstrand.replace(i,1,"C",1);
            else if(strand.at(i) == 'T') rcstrand.replace(i,1,"A",1);
        }
	// Note that rcseq now contains the complementary RNA strand! Only, the order is still incorrect. Thus, let us reorder:
        string temp;
        int length;
        char tmpchar;
        string charstr;
        length = strand.length();
        temp = rcstrand.at(length-1);
        for(int i = 1; i<length; i++)
        {
            tmpchar = rcstrand.at(length-i-1);
            charstr = tmpchar;
       	    temp.append(charstr);
        }
        rcseq = temp; // Invert direction, now it is in 5'-3' direction
	symmcheck = 1;
	for(int i = 0; i<length; i++) symmcheck *= (seq.at(i) == rcseq.at(i)); // This is the symmetry check
        combseq = seq;
        combseq.append("XXXXX");
        combseq.append(rcseq);
	termAUcount = 0;
	if(seq.at(0)=='A' || seq.at(0)=='B' || seq.at(0)=='T' || seq.at(0)=='U') termAUcount++;
	if(rcseq.at(0)=='A' || rcseq.at(0)=='B' || rcseq.at(0)=='T' || rcseq.at(0)=='U') termAUcount++;
	numstrand[0] = new short int[seq.length()];
	numstrand[1] = new short int[rcseq.length()];
	numstrand[2] = new short int[combseq.length()];
	translate(seq,numstrand[0]);
	translate(rcseq,numstrand[1]);
	translate(combseq,numstrand[2]);
    }
    
    Sequence(int id, string str, double energ, bool dummy, double energRef, double dh, double ds, bool dummy2) // RNA/RNA Sequence constructor
    {
	seqID = id-1; // The list starts at 1, but C++ arrays start with a zeroth element. This seems more intuitive, even if it may cause confusion.
	seqID += int(dummy); // prevent an 'unused' warning, dummy is used only to distinguish RNA/RNA from DNA/DNA constructor and has no purpose other than that.
	seqID += int(dummy2); // Same reasoning
	deltaG = energ;
	deltaGref = energRef;
	deltaH = dh;
	deltaS = ds;
        seq = str;
	string strand = str;
	string rcstrand = str;
        for(unsigned int i = 0; i<strand.length(); i++)
        {
            if(strand.at(i) == 'B') rcstrand.replace(i,1,"U",1);
            else if(strand.at(i) == 'D') rcstrand.replace(i,1,"H",1);
            else if(strand.at(i) == 'H') rcstrand.replace(i,1,"D",1);
            else if(strand.at(i) == 'U') rcstrand.replace(i,1,"B",1);
        }
	// Note that rcseq now contains the complementary RNA strand! Only, the order is still incorrect. Thus, let us reorder:
        string temp;
        int length;
        char tmpchar;
        string charstr;
        length = strand.length();
        temp = rcstrand.at(length-1);
        for(int i = 1; i<length; i++)
        {
            tmpchar = rcstrand.at(length-i-1);
            charstr = tmpchar;
       	    temp.append(charstr);
        }
        rcseq = temp; // Invert direction, now it is in 5'-3' direction
	symmcheck = 1;
	for(int i = 0; i<length; i++) symmcheck *= (seq.at(i) == rcseq.at(i)); // This is the symmetry check
        combseq = seq;
        combseq.append("XXXXX");
        combseq.append(rcseq);
	termAUcount = 0;
	if(seq.at(0)=='A' || seq.at(0)=='B' || seq.at(0)=='T' || seq.at(0)=='U') termAUcount++;
	if(rcseq.at(0)=='A' || rcseq.at(0)=='B' || rcseq.at(0)=='T' || rcseq.at(0)=='U') termAUcount++;
	numstrand[0] = new short int[seq.length()];
	numstrand[1] = new short int[rcseq.length()];
	numstrand[2] = new short int[combseq.length()];
	translate(seq,numstrand[0]);
	translate(rcseq,numstrand[1]);
	translate(combseq,numstrand[2]);
    }

    ~Sequence() // Destructor
    {
	delete [] numstrand[0];
	delete [] numstrand[1];
	delete [] numstrand[2];
    }

    string getseq() const
    { return seq; }

    string getrcseq() const
    { return rcseq; }

    string getcombseq() const
    { return combseq; }

    double getdeltaG() const
    { return deltaG; }

    double getdeltaGref() const
    { return deltaGref; }

    double getdeltaH() const
    { return deltaH; }

    double getdeltaS() const
    { return deltaS; }

    int gettermAUcount() const
    { return termAUcount; }

    bool getsymmcheck() const
    { return symmcheck; }

    int gettype() const
    {
	if(numstrand[2][0] < 4 && numstrand[2][combseq.length()-1] < 4) return 0;
	else if(numstrand[2][0] > 3 && numstrand[2][combseq.length()-1] > 3) return 2;
	else return 1;
    }

// Set configuration
    void setConfig(string str)
    {
	cout << "The setConfig function has been disabled!";
	exit(1);
    }
						    
}; // here ends the Sequence class declaration

    void initseqs(Sequence*); // Create a list of sequences

#endif
